/*
 * Copyright (c) 2012 Petr Cerny.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "includes.h"

#include "fips.h"

#include "cipher.h"
#include "dh.h"
#include "digest.h"
#include "kex.h"
#include "key.h"
#include "mac.h"
#include "log.h"
#include "xmalloc.h"

#include <string.h>
#include <openssl/crypto.h>

/* import from dh.c */
extern int dh_grp_min;

static int fips_state = -1;

static int
fips_check_required_env(void)
{
	int fips_required = 0;
	char *env = getenv(SSH_FORCE_FIPS_ENV);

	if (env) {
		errno = 0;
		fips_required = strtol(env, NULL, 10);
		if (errno) {
			debug("bogus value in the %s environment variable, ignoring\n"
			    , SSH_FORCE_FIPS_ENV);
			fips_required = 0;
		} else
			fips_required = 1;
	}
	return fips_required;
}

int
fips_mode(void)
{
	if (-1 == fips_state) {
		fips_state = FIPS_mode();
		if (fips_state)
			debug("FIPS mode initialized");
		else {
			if (fips_check_required_env()) {
				debug("FIPS mode requested through the environment variable '%s'"
				    , SSH_FORCE_FIPS_ENV);
				if (!FIPS_mode_set(1))
					fatal("Unable to enter FIPS mode as requested through the environment variable '%s'"
					    , SSH_FORCE_FIPS_ENV);
				fips_state = 1;
			}
		}
	}
	return fips_state;
}

int
fips_correct_dgst(int digest)
{
	int fips;
	int rv = -1;

	fips = fips_mode();
	switch (fips) {
		case 0:
			rv = digest;
			break;
		case 1:
			switch (digest) {
				case SSH_DIGEST_MD5:
				case SSH_DIGEST_RIPEMD160:
					debug("MD5/RIPEMD160 digests not allowed in FIPS 140-2 mode"
					    "using SHA-256 instead.");
					rv = SSH_DIGEST_SHA256;
					break;
				default:
					rv = digest;
					break;
			}
			break;
		default:
			/* should not be reached */
			fatal("Fatal error: incorrect FIPS mode '%i' at %s:%u",
			    fips, __FILE__, __LINE__);
	}

	return rv;
}

/*
 * filter out FIPS disallowed algorithms
 * *crypto MUST be free()-able - it is assigned newly allocated memory and
 * the previous one is freed
 *
 * returns zero if all algorithms were rejected, non-zero otherwise
 */
int
fips_filter_crypto(char **crypto, fips_filters filter)
{
	char *token, *tmp, *tmp_sav, *new;
	int plus = 0;
	int valid;
	int comma = 0;
	int empty = 1;
	size_t len;

	tmp = tmp_sav = xstrdup(*crypto);

	len = strlen(tmp) + 1;
	new = xcalloc(1, len);

	if ('+' == *tmp) {
		plus = 1;
		tmp++;
	}

	while ((token = strsep(&tmp, ",")) != NULL) {
		switch(filter) {
			case FIPS_FILTER_CIPHERS:
				valid = ciphers_valid(token);
				if (!valid)
					debug("Cipher '%s' is not allowed in FIPS mode",
					    token);
				break;
			case FIPS_FILTER_MACS:
				valid = mac_valid(token);
				if (!valid)
					debug("MAC '%s' is not allowed in FIPS mode",
					    token);
				break;
			case FIPS_FILTER_KEX_ALGS:
				valid = kex_names_valid(token);
				if (!valid)
					debug("KEX '%s' is not allowed in FIPS mode",
					    token);
				break;
			default:
				/* should not be reached */
				fatal("Fatal error: incorrect FIPS filter '%i' requested at %s:%u",
				    filter, __FILE__, __LINE__);
		}

		if (valid) {
			empty = 0;
			if (plus) {
				strlcat(new, "+", len);
				plus = 0;
			}
			if (comma)
				strlcat(new, ",", len);
			else
				comma = 1;
			strlcat(new, token, len);
		}
	}

	/* free tmp and re-allocate shorter buffer for result if necessary */
	free(tmp_sav);
	free(*crypto);
	*crypto = new;

	return (!empty);
}

int
fips_dgst_min(void)
{
	int fips;
	int dgst;

	fips = fips_mode();
	switch (fips) {
		case 0:
			dgst = SSH_DIGEST_MD5;
			break;
		case 1:
			dgst = SSH_DIGEST_SHA1;
			break;
		default:
			/* should not be reached */
			fatal("Fatal error: incorrect FIPS mode '%i' at %s:%u",
			    fips, __FILE__, __LINE__);
	}
	return dgst;
}

int
fips_dh_grp_min(void)
{
	int fips;
	int dh;

	fips = fips_mode();
	switch (fips) {
		case 0:
			dh = dh_grp_min;
			break;
		case 1:
			dh = DH_GRP_MIN_FIPS;
			break;
		default:
			/* should not be reached */
			fatal("Fatal error: incorrect FIPS mode '%i' at %s:%u",
			    fips, __FILE__, __LINE__);
	}
	return dh;
}

